from __future__ import annotations

import json

from xattrs import frozen
from xattrs.preconf.json import to_json


def test_to_json():
    @frozen
    class Person:
        name: str
        age: int

    # Decoding a Person object
    obj = {"name": "John", "age": 25}
    person = Person(**obj)

    assert repr(person) == "Person(name='John', age=25)"

    assert to_json(person) == json.dumps(obj)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._typing import T
from xattrs.typing import CaseConvention, CaseConverter

import re


def identity(val: T) -> T:
    return val


def to_lower(value: str) -> str:
    return value.lower()


def to_upper(value: str) -> str:
    return value.upper()


def to_capital(value: str) -> str:
    return value.capitalize()


def to_camel(value: str) -> str:
    pascal = to_pascal(value)
    if len(pascal) <= 1:
        return pascal.lower()
    return pascal[0].lower() + pascal[1:]


def to_pascal(value: str) -> str:
    if "-" in value:
        value = value.replace("-", "_")

    expr = re.compile(r"(?:^|_)(.)")
    return expr.sub(lambda m: m.group(1).lower().title(), value)


def to_snake(value: str) -> str:
    if "-" in value:
        value = "".join(
            w.lower().capitalize() for w in value.replace("_", "-").split("-")
        )

    expr0 = re.compile(r"(.)([A-Z][a-z]+)")
    expr1 = re.compile(r"([a-z0-9])([A-Z])")
    repl = r"\1_\2"
    return expr1.sub(repl, expr0.sub(repl, value)).lower()


def to_kebab(value: str) -> str:
    return to_snake(value).replace("_", "-")


def to_secret(value: str, marker: str = "*", length: int | None = 6) -> str:
    length = int(length or len(value))
    return "".join(str(marker) for _ in range(length))


def to_const(value: str) -> str:
    return to_snake(value).upper()


_CASE_CONVERTER_MAPPING: dict[CaseConvention, CaseConverter] = {
    "lowercase": to_lower,
    "uppercase": to_upper,
    "UPPERCASE": to_upper,
    "capitalcase": to_capital,
    "Capitalcase": to_capital,
    "snake_case": to_snake,
    "camelCase": to_camel,
    "kebab-case": to_kebab,
    "CONST_CASE": to_upper,
    "PascalCase": to_pascal,
}



# SPDX-License-Identifier: MIT
from __future__ import annotations

from typing import NoReturn
from xattrs._compat.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Final,
    Mapping,
    cast,
)
from xattrs._typing import (
    Dispatchable,
    Hook,
    SingleDispatchCallable,
    T,
    UnarySingleDispatchCallable,
)

from functools import singledispatch

from xattrs.abc import AbstractDispatcher
from xattrs.exceptions import HookNotFoundError


def _dispatch_not_found(_, **kwargs) -> NoReturn:
    raise HookNotFoundError("Hook not found. This raised by not found fallback hook.")


_unary_hook_registry: dict[type[T], Callable[[Any], T]] = {
    # builtin types
    # error/exception types
    # warning types
}

# slice: lambda x: slice(x),
_error_hook = {}
_warning_hook = {}

# generic types:
# tuple: lambda x: tuple(x),
# list: lambda x: list(x),
# set: lambda x: set(x),
# dict: lambda x: dict(x),
# frozenset: lambda x: frozenset(x)


class ConcreteDispatch(AbstractDispatcher[T]):
    """Dispatch for concrete types."""

    _registry: dict[type[T], Hook[T]] = _unary_hook_registry  # pyright: ignore[reportAssignmentType] # fmt: skip

    # def register(self, cls: Dispatchable[T], func: Hook[T]) -> Hook[T]:
    #     self._registry[cls] = func
    #     return func

    def dispatch(self, cls: type[T]) -> Hook[T]:
        hook = self._registry.get(cls)
        if hook is None:
            raise HookNotFoundError(f"Hook for {cls} not found.")
        return hook


class SingleDispatcher(AbstractDispatcher[T]):
    """Single dispatch dispatcher

    This is a dispatcher that uses the ``singledispatch`` function for ``covariant`` types.
    """

    _registry: ClassVar[...]

    def __init__(self) -> None:
        self._singledispatch = singledispatch(_dispatch_not_found)

    def register(self, cls: type[T], func: Hook[T]) -> Hook[T]:
        return self._singledispatch.register(cls, func)

    def dispatch(self, cls: type[T]) -> Hook[T]:
        return self._singledispatch.dispatch(cls)


class GenericDispatcher:
    """Generic dispatcher

    This is a dispatcher that uses the ``singledispatch`` function for ``contravariant`` types.
    """

    _registry: ClassVar



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._typing import DataclassInstance

from dataclasses import (
    KW_ONLY,
    MISSING,
    Field,
    FrozenInstanceError,
    InitVar,
    asdict,
    astuple,
    field,
    fields,
    is_dataclass,
    replace,
)
from dataclasses import dataclass as _dataclass

import optree

from xattrs.constants import METANAME

# Keep the same with the `dataclasses` module of std lib, except `make_dataclass`.
# `make_dataclass` is not a full replacement  in `xattrs`.
__all__ = [
    "KW_ONLY",
    "MISSING",
    "Field",
    "FrozenInstanceError",
    "InitVar",
    "asdict",
    "astuple",
    "dataclass",
    "field",
    "fields",
    "is_dataclass",
    # "make_dataclass",
    "replace",
]


def _dataclass_flatten(ins: DataclassInstance):
    return (ins.__dict__, fields(ins))  # type: ignore[reportArgumentType]


def dataclass(  # noqa: PLR0913
    cls,
    /,
    *,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
    match_args=True,
    kw_only=False,
    slots=False,
):
    def wrap(_cls):
        dataclass_cls = _dataclass(
            _cls,
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
            match_args=match_args,
            kw_only=kw_only,
            slots=slots,
        )

        def _dataclass_unflatten(metadata, children) -> DataclassInstance:
            return dataclass_cls(**metadata)

        optree.register_pytree_node(
            dataclass_cls, _dataclass_flatten, _dataclass_unflatten, namespace=METANAME
        )

        return dataclass_cls

    if cls is None:
        return wrap

    return wrap(cls)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Hashable,
    Literal,
    Protocol,
)
from xattrs._typing import T, T_co, T_contra

from dataclasses import Field

from attrs import Attribute

StructAs = Literal["dict", "tuple"]

# easy to remember?
CaseConvention = Literal[
    "lowercase",
    "UPPERCASE",
    "uppercase",
    "Capitalcase",
    "capitalcase",
    "snake_case",
    "kebab-case",
    "CONST_CASE",
    "camelCase",
    "PascalCase",
    # "Ada_Case", # NotImplemented
    # "COBOL-CASE", # NotImplemented
    # "Train-Case", # NotImplemented
]
CaseConverter = Callable[[str], str]
KeyConverter = Callable[[str], str]

FilterBuiltins = Literal[
    "exclude_if_default",
    "exclude_if_none",
    "exclude_if_false",
]
if TYPE_CHECKING:
    FilterCallable = (
        Callable[[Attribute[T], T], bool]
        | Callable[[Field[T], T], bool]
        | Callable[[Hashable, T], bool]
    )
else:
    FilterCallable = Any

UnknownFields = Literal["ignore", "allow", "deny"]


class SerializeFunc(Generic[T_contra, T_co], Protocol):
    def __call__(self, value: T_contra, **kwargs: Any) -> T_co: ...


class DeserializeFunc(Generic[T, T_contra], Protocol):
    def __call__(self, value: T_contra, cls: type[T], **kwargs: Any) -> T: ...



from __future__ import annotations

from typing import Any, Callable

from dataclasses import dataclass
from dataclasses import field as dataclass_field

import pytest
from attr import field as attrs_filed
from attrs import define
from hypothesis import given
from hypothesis import strategies as st


class TestGenSerializerHelpers: ...



# SPDX-License-Identifier: MIT
# ruff: noqa: F401, UP035
# mypy: disable-error-code="assignment"
from typing import *  # noqa: F403 # pyright: ignore[reportWildcardImportFromLibrary]
from typing_extensions import __all__

from sys import version_info

# Overwrite the original `typing` module from `typing_extensions`
# for typing compatibility

if version_info < (3, 12):
    from typing_extensions import TypeAliasType as TypeAliasType
    from typing_extensions import override as override

if version_info < (3, 11):
    from typing_extensions import Any as Any  # behavior changed in 3.11
    from typing_extensions import LiteralString as LiteralString
    from typing_extensions import NamedTuple as NamedTuple  # behavior changed in 3.11
    from typing_extensions import Never as Never
    from typing_extensions import (
        NewType as NewType,  # behavior changed in 3.10, performance improved in 3.11
    )
    from typing_extensions import NotRequired as NotRequired
    from typing_extensions import Required as Required
    from typing_extensions import Self as Self
    from typing_extensions import TypedDict as TypedDict  # behavior changed in 3.11
    from typing_extensions import TypeVarTuple as TypeVarTuple
    from typing_extensions import Unpack as Unpack
    from typing_extensions import assert_never as assert_never
    from typing_extensions import assert_type as assert_type
    from typing_extensions import clear_overloads as clear_overloads
    from typing_extensions import dataclass_transform as dataclass_transform
    from typing_extensions import final as final  # behavior changed in 3.11
    from typing_extensions import get_overloads as get_overloads
    from typing_extensions import (
        get_type_hints as get_type_hints,  # behavior changed in 3.11
    )
    from typing_extensions import overload as overload  # behavior changed in 3.11
    from typing_extensions import reveal_type as reveal_type

if version_info < (3, 10):
    from typing_extensions import Annotated as Annotated
    from typing_extensions import Callable as Callable
    from typing_extensions import Concatenate as Concatenate
    from typing_extensions import Generic as Generic
    from typing_extensions import ParamSpec as ParamSpec
    from typing_extensions import ParamSpecArgs as ParamSpecArgs
    from typing_extensions import ParamSpecKwargs as ParamSpecKwargs
    from typing_extensions import TypeAlias as TypeAlias
    from typing_extensions import TypeGuard as TypeGuard
    from typing_extensions import is_typeddict as is_typeddict

if version_info < (3, 9, 1):
    # from docs.python.org/3/library/typing.html#typing.Literal
    # _Changed in version 3.9.1_: `Literal` now de-duplicates parameters.
    from typing_extensions import Literal as Literal



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._uni import _ATTRS_ATTRS, _DATACLASS_FIELDS


def has(cls: type) -> bool:
    """Check whether **cls** is a class with `attrs` attributes or `dataclass` fields."""
    return hasattr(cls, _ATTRS_ATTRS) or hasattr(cls, _DATACLASS_FIELDS)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import Callable
from xattrs._typing import AttrsInstance, DataclassInstance, T_interm, T_proto

from xattrs._helpers import _identity
from xattrs._serde import _maybe_serde
from xattrs.base import BaseSerializer
from xattrs.converters import _CASE_CONVERTER_MAPPING


class Serializer(BaseSerializer[T_interm, T_proto]): ...


_KeySerializer = Callable
_ValueSerializer = Callable



# SPDX-License-Identifier: MIT
"""
Compatibility module for those internal functions or classes not exposed from
the ``typing`` module even not exposed from the ``typing_extensions`` module.
"""

from typing_extensions import _AnnotatedAlias



# SPDX-License-Identifier: MIT
from __future__ import annotations

import types
from xattrs._compat.typing import Any, AnyStr, Callable, TypeVar, Union
from xattrs.typing import CaseConvention, DeserializeFunc, SerializeFunc

from datetime import datetime
from functools import partial

from xattrs._struct_funcs import asdict_shallow
from xattrs.deserializer import Deserializer
from xattrs.serializer import Serializer

__all__ = ("dump", "dumps", "from_msgpack", "load", "loads", "to_msgpack")


MsgPackable = Union[dict, list, tuple, str, int, float, bool, types.NoneType]  # type: ignore[type-arg]


class MsgPackSerializer(Serializer[bytes, MsgPackable]): ...


class MsgPackDeserializer(Deserializer[bytes, MsgPackable]): ...



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import TYPE_CHECKING, Callable

from attr import dataclass
from attrs import define, evolve, field, fields, frozen, mutable

from xattrs._serde import serde
from xattrs._struct_funcs import asdict, astree, astuple

if TYPE_CHECKING:
    from xattrs._typing import Decorator, P, R_co


__all__ = (
    "asdict",
    "astree",
    "astuple",
    "dataclass",
    "define",
    "derive",
    "evolve",
    "field",
    "fields",
    "frozen",
    "mutable",
    "replace",
    "serde",
)

replace = evolve


def derive(*traits: Decorator[P, R_co]) -> Decorator[P, R_co]:
    """
    Derive a new class or a new `Callable` with the given traits.

    Simpely compose multiple decorators into one for now.
    """

    def decorator(wrapped: Callable[P, R_co]) -> Callable[P, R_co]:
        for trait in traits:
            wrapped = trait(wrapped)
        return wrapped

    return decorator



from __future__ import annotations

from typing import Any

from collections import OrderedDict
from dataclasses import dataclass

import pytest
from attrs import frozen
from hypothesis import given
from hypothesis import strategies as st

from xattrs._struct_funcs import astuple

MAPPING_TYPES = (dict, OrderedDict)
SEQUENCE_TYPES = (list, tuple)


@pytest.fixture(scope="class")
def A():
    """
    Return an attrs class
    """

    @frozen
    class A:
        x: Any
        y: Any

    return A


@pytest.fixture(scope="class")
def D():
    """
    Return a dataclass class
    """

    @dataclass
    class D:
        x: Any
        y: Any

    return D


class TestAsTuple:
    """
    Tests for `astuple`.
    """

    @given(seq=st.sampled_from(SEQUENCE_TYPES))
    def test_attrs(self, A, seq):
        acutal = astuple(A(A(1, 2), A(3, 4)), tuple_factory=seq)
        expected = seq((seq((1, 2)), seq((3, 4))))
        assert acutal == expected

    @given(seq=st.sampled_from(SEQUENCE_TYPES))
    def test_dataclasses(self, D, seq):
        acutal = astuple(D(D("1", 2), D(3, "4")), tuple_factory=seq)
        expected = seq((seq(("1", 2)), seq((3, "4"))))
        assert acutal == expected

    def test_mixed_attrs_dataclass(self, A, D):
        actual = (("1", "2"), (3, 4))
        expected = astuple(A(A("1", "2"), D(3, 4)))
        assert actual == expected

        actual = (("1", 2), (3, "4"))
        expected = astuple(D(A("1", 2), D(3, "4")))
        assert actual == expected

    def test_mapping(
        self,
        A,
        D,
    ):
        """
        contains mapping
        """
        actual = astuple(D(1, {"x": 2, "y": 3}))
        expected = (1, {"x": 2, "y": 3})
        assert actual == expected

        actual = astuple(
            A(
                1,
                (A(2, 3), D(4, 5), {"x": "6", "y": "7"}),
            )
        )
        expected = (
            1,
            ((2, 3), (4, 5), {"x": "6", "y": "7"}),
        )
        assert actual == expected



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._typing import T_interm, T_proto

from xattrs.base import BaseDeserializer


class Deserializer(BaseDeserializer[T_proto, T_interm]): ...



# SPDX-License-Identifier: MIT

import types
from xattrs._compat.typing import Any, AnyStr, Callable, TypeVar, Union
from xattrs.typing import DeserializeFunc, SerializeFunc

import json
from datetime import datetime

from xattrs._struct_funcs import asdict_shallow
from xattrs.deserializer import Deserializer
from xattrs.serializer import Serializer

__all__ = ["from_json", "to_json"]

# Supports the following objects and types by default:
#
# +-------------------+---------------+
# | Python            | JSON          |
# +===================+===============+
# | dict              | object        |
# +-------------------+---------------+
# | list, tuple       | array         |
# +-------------------+---------------+
# | str               | string        |
# +-------------------+---------------+
# | int, float        | number        |
# +-------------------+---------------+
# | True              | true          |
# +-------------------+---------------+
# | False             | false         |
# +-------------------+---------------+
# | None              | null          |
# +-------------------+---------------+
#

T = TypeVar("T")

Jsonable = Union[dict, list, tuple, str, int, float, bool, types.NoneType]  # type: ignore[type-arg,valid-type]


class JsonDeserializer(Deserializer[AnyStr, Jsonable]):
    """JSON deserializer."""

    def _datetime_from_isoformat(self, value: str) -> datetime:
        """Convert the value to a Python object."""
        return datetime.fromisoformat(value)

    def loads(self, data: AnyStr, **kwargs: Any) -> Any:
        """Deserialize the JSON string to an object."""
        return json.loads(data)

    def from_json(
        self,
        s: AnyStr,
        cls: type[T],
        /,
        **kw,
    ) -> T:
        """Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance
        containing a JSON document) to a Python object.
        """
        # deserializer = deserializer or default_json_deserializer
        return json.loads(s, cls, **kw)


class JsonSerializer(Serializer[Jsonable, str]):
    """JSON serializer."""

    def _datetime_to_isoformat(self, value: datetime) -> str:
        """Convert the value to an intermediate data types."""
        return value.isoformat()

    def dumps(self, obj: Jsonable, **kwargs: Any) -> str:
        """Serialize the object to a JSON-formatted string."""
        return json.dumps(obj)

    def to_json(
        self,
        obj: Any,
        *,
        key_serializer: Callable[[str], str] | None = None,
        value_serializer: SerializeFunc | None = None,
        dumps: Callable | None = None,
        **kwargs: Any,
    ) -> str:
        """Serialize ``obj`` to a JSON-formatted ``str``."""
        value_serializer = value_serializer or self
        dumps = dumps or json.dumps
        return dumps(
            asdict_shallow(
                obj, key_serializer=key_serializer, value_serializer=value_serializer
            ),
            **kwargs,
        )


json_serializer = JsonSerializer()
json_deserializer = JsonDeserializer()

to_json = json_serializer.to_json
from_json = json_deserializer.from_json
dumps = json_serializer.dumps
loads = json_deserializer.loads
# dump = json_serializer.dump
# load = json_deserializer.load



# SPDX-License-Identifier: MIT
from __future__ import annotations

from types import MappingProxyType
from typing import Literal
from xattrs._compat.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Concatenate,
    Generic,
    LiteralString,
    Mapping,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    TypeVarTuple,
    Union,
)

from dataclasses import Field

from attr._make import _CountingAttr  # noqa: PLC2701
from attrs import Attribute

_T = TypeVar("_T")


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


class AttrsInstance(Protocol):
    __attrs_attrs__: ClassVar[tuple[str, Attribute[Any]]]


T_attrs = TypeVar("T_attrs", bound=type[AttrsInstance])
T_dataclass = TypeVar("T_dataclass", bound=type[DataclassInstance])

if TYPE_CHECKING:
    DataclassLike = TypeVar("DataclassLike", DataclassInstance, AttrsInstance)
    FieldLike = Field[_T] | Attribute[_T] | _CountingAttr

    class _DataclassParams:
        init: bool
        repr: bool
        eq: bool
        order: bool
        unsafe_hash: bool
        frozen: bool
        match_args: bool
        kw_only: bool
        slots: bool
        weakref_slot: bool

    class _AttrsParams:
        init: bool
        repr: bool
        eq: bool
        order: bool
        unsafe_hash: bool
        frozen: bool
        match_args: bool
        kw_only: bool
        slots: bool
        weakref_slot: bool

else:
    DataclassLike = Any
    FieldLike = Any

    class _DataclassParams: ...

    class _AttrsParams: ...


S = TypeVar("S", bound=str)
A = TypeVar("A")
B = TypeVar("B")

K = TypeVar("K")
V = TypeVar("V")

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)

P = ParamSpec("P")
P1 = ParamSpec("P1")
P2 = ParamSpec("P2")
R_co = TypeVar("R_co", covariant=True)
Ts = TypeVarTuple("Ts")

# T_ls = TypeVar("T_ls", LiteralString)

Decorator: TypeAlias = Callable[[Callable[P, R_co]], Callable[P, R_co]]

if TYPE_CHECKING:
    PredCallable: TypeAlias = Callable[P, bool]
    UnaryCallable: TypeAlias = Callable[Concatenate[T, P], R_co]  # Support at least 1 argument # fmt: skip
    BinaryCallable: TypeAlias = Callable[Concatenate[T1, T2, P], R_co]  # Support at least 2 arguments # fmt: skip
    CallableFacatory: TypeAlias = Callable[P1, Callable[P2, R_co]]

    TreeCallable: TypeAlias = Callable[[tuple[Any]], R_co]
    GenericCallable: TypeAlias = Callable[..., R_co]

else:
    PredCallable = Any
    UnaryCallable = Any
    BinaryCallable = Any
    CallableFacatory = Any

    TreeCallable = Any
    GenericCallable = Any


# ---- Framework specific types ----
# Generic for final protocol type
T_proto = TypeVar("T_proto", covariant=True)
# interchange datatype / intermediary / bridge
# Protocol-aware type
# Protocol compatible Python type
T_interm = TypeVar("T_interm")

if TYPE_CHECKING:
    HookPredicate: TypeAlias = Callable[Concatenate[type[T], P], bool]
    HookFacatory: TypeAlias = Callable[Concatenate[type[T], P], Callable[..., T]]
    ConstructHook: TypeAlias = BinaryCallable[type[T1], T2, ..., T1]
    DeconstructHook: TypeAlias = UnaryCallable[type[T], ..., T]
    Hook = Union[ConstructHook[T, Any], DeconstructHook[T]]
    Dispatchable = Union[type[T], HookPredicate[T, ...]]
else:
    HookPredicate = Any
    HookFacatory = Any
    ConstructHook = Any
    DeconstructHook = Any
    Hook = Any
    Dispatchable = Any


class SingleDispatchCallable(Generic[T]):
    """polymorphism on single parameter"""

    registry: MappingProxyType[Any, GenericCallable[T]]

    def register(
        self, cls: type[T], func: GenericCallable[T]
    ) -> GenericCallable[T]: ...
    def dispatch(self, cls: type[T]) -> GenericCallable[T]: ...
    def _clear_cache(self) -> None: ...
    def __call__(self, /, *args: Any, **kwargs: Any) -> T: ...


class UnarySingleDispatchCallable(SingleDispatchCallable[T]):
    # fmt: off
    def register(self, cls: type[T], func: UnaryCallable[T, ..., T]) -> UnaryCallable[T, ..., T]: ...
    def dispatch(self, cls: type[T]) -> UnaryCallable[T, ..., T]: ...
    # fmt: on


class TreeSingleDispatchCallable(SingleDispatchCallable[T]):
    def register(self, cls: type[T], func: TreeCallable[T]) -> TreeCallable[T]: ...
    def dispatch(self, cls: type[T]) -> TreeCallable[T]: ...


class PredicateSingleDispatchCallable:
    # fmt: off
    def register(self, pred: PredCallable[P], func: GenericCallable[T]) -> GenericCallable[T]: ...
    def dispatch(self, pred: PredCallable[P]) -> GenericCallable[T]: ...
    # fmt: on



# isort:skip_file


def test_imports():
    from xattrs import dataclass, define, frozen, field
    from xattrs import asdict, astuple, astree
    from xattrs import derive
    from xattrs import serde
    from xattrs import evolve



# SPDX-License-Identifier: MIT
from __future__ import annotations


class XAttrsException(Exception):
    """
    Base class for all exceptions raised by ``xattrs``.
    """


class DispatchError(XAttrsException):
    """
    Raised when error occurs during excution of dispatching.
    """


class SingleDispatchError(DispatchError):
    """
    Raised when get a hook or apply a hook through single dispatch branch.
    """


class HookNotFoundError(DispatchError):
    """
    Raised when a hook for a type is not found from dispatch register.
    """



# SPDX-License-Identifier: MIT
from __future__ import annotations

from functools import partial

from optree import register_pytree_node, tree_flatten, tree_unflatten

__all__ = ["flatten", "register_de_node", "register_ser_node", "unflatten"]

_SER_NAMESPACE = "__attrs_ser__"
_DE_NAMESPACE = "__attrs_deser__"

_FLATTEN_ATTR = "__attrs_flatten__"
_UNFLATTEN_ATTR = "__attrs_unflatten__"

register_ser_node = partial(register_pytree_node, namespace=_SER_NAMESPACE)
register_de_node = partial(register_pytree_node, namespace=_DE_NAMESPACE)

# flatten = partial(tree_flatten, namespace=_TREE_NAMESPACE)
# unflatten = tree_unflatten



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import Any, Literal, cast
from xattrs._typing import FieldLike


def keep_include(field: Any, value: Any) -> Literal[True]:
    return True


def exclude_if_default(field: FieldLike[Any], value: Any) -> bool:
    if field.default is not None:
        if not callable(field.default):
            # if not default return True to include
            return cast(bool, value != field.default)
        else:
            raise NotImplementedError("Callable default values are not supported yet.")
    else:
        # no default value
        return True


# TODO: Add test cases
def exclude_if_none(field: FieldLike[Any], value: Any) -> bool:
    return value is not None


def exclude_if_false(field: Any, value: Any) -> bool:
    """
    >>> True if 0 else False
    >>> False
    >>> True if [] else False
    >>> False

    >>> bool(0)
    >>> False
    >>> bool([])
    >>> False
    """
    return bool(value)



# SPDX-License-Identifier: MIT
"""
Read more about Metadata: https://www.attrs.org/en/stable/extending.html#metadata
"""
from __future__ import annotations

from xattrs._compat.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Mapping,
    TypedDict,
    TypeGuard,
    cast,
    overload,
)
from xattrs._typing import T
from xattrs.typing import CaseConvention, CaseConverter, FilterCallable, StructAs

from dataclasses import Field, asdict, dataclass

from attr._make import _CountingAttr

from xattrs.converters import _CASE_CONVERTER_MAPPING, identity
from xattrs.filters import exclude_if_default as exclude_if_default_filter
from xattrs.filters import exclude_if_false as exclude_if_false_filter
from xattrs.filters import keep_include

if TYPE_CHECKING:
    from attrs import Attribute


@dataclass(slots=True)
class _Metadata(Generic[T]):
    name: str | None = None
    rename: CaseConvention | CaseConverter | None = None
    # rename_ser: str | None = None
    # rename_de: str | None = None

    exclude: bool | None = None
    # exclude_ser: bool = False
    # exclude_de: bool = False
    exclude_if: FilterCallable[T] | None = None
    exclude_if_default: bool | None = None
    exclude_if_false: bool | None = None

    converter_to: Callable[[T], Any] | None = None  # same as converter
    converter_from: Callable[[Any], T] | None = None

    kind: StructAs | None = None
    flatten: bool = False  # inspired from pyserde

    def __post_init__(self) -> None:
        # check conflicts for exclude params
        num_excludes = sum(
            1 if x is not None else 0
            for x in (
                self.exclude,
                self.exclude_if,
                self.exclude_if_default,
                self.exclude_if_false,
            )
        )
        if num_excludes > 1:
            raise ValueError(
                "Only one of 'exclude', 'exclude_if', 'exclude_if_default', "
                "'exclude_if_false' values can be set."
            )

    @overload
    def __ror__(self, value: Field[T]) -> Field[T]: ...
    @overload
    def __ror__(self, value: _CountingAttr) -> _CountingAttr: ...
    def __ror__(self, value: Any) -> Field | _CountingAttr:  # type: ignore[type-arg]
        if isinstance(value, Field):
            metadata = {**value.metadata, **asdict(self)}
            return Field(
                default=cast(T, value.default),
                default_factory=cast(Callable[[], T], value.default_factory),
                init=value.init,
                repr=value.repr,
                hash=value.hash,
                compare=value.compare,
                metadata=metadata,
                kw_only=cast(bool, value.kw_only),
            )
        elif isinstance(value, _CountingAttr):
            # avoid to increase cls counter
            value.metadata |= asdict(self)
            return value
        else:
            raise TypeError(
                f"Expected an 'DataclassLike' instance, got {type(value)!r}"
            )


class FilterConf(TypedDict):
    exclude: bool
    exclude_if: FilterCallable | None  # type: ignore[type-arg]
    exclude_if_default: bool | None
    exclude_if_false: bool | None


def _has_filter_params(meta: Mapping) -> TypeGuard[FilterConf]:  # type: ignore[type-arg]
    return any(
        (
            meta.get("exclude") is not None,
            meta.get("exclude_if") is not None,
            meta.get("exclude_if_default") is not None,
            meta.get("exclude_if_false") is not None,
        )
    )


# FIXME(@lqhuang): update to pipeline scope filter rather than replace scope filter
def _gen_field_filter(
    attribute: Field[T] | Attribute[T], scope_filter: FilterCallable[T] | None = None
) -> FilterCallable[T]:
    """Gen a filter function by merging the field and scope filter."""
    _meta = attribute.metadata
    if not _has_filter_params(_meta):
        return scope_filter or keep_include
    elif _meta.get("exclude"):
        return lambda _, __: False  # return False to exclude this field
    elif (exclude_if_filter := _meta.get("exclude_if")) is not None:
        return exclude_if_filter
    elif _meta.get("exclude_if_default"):
        return exclude_if_default_filter
    elif _meta.get("exclude_if_false"):
        return exclude_if_false_filter
    else:
        return keep_include  # only right for replacing scope filter


class RenameConf(TypedDict):
    name: str | None
    rename: CaseConvention | CaseConverter | None


def _has_key_serializer_params(meta: Mapping[str, Any]) -> TypeGuard[RenameConf]:
    return any(
        (
            meta.get("name") is not None,
            meta.get("rename") is not None,
        )
    )


def _gen_field_key_serializer(
    attribute: Field[T] | Attribute[T],
    scope_key_serializer: CaseConverter | None = None,
) -> CaseConverter:
    _meta = attribute.metadata
    if not _has_key_serializer_params(_meta):
        return scope_key_serializer or identity
    if (_name := _meta.get("name")) is not None:
        return lambda x: _name
    elif (_rename := _meta.get("rename")) is not None:
        if isinstance(_rename, str):
            try:
                return _CASE_CONVERTER_MAPPING[_rename]
            except KeyError:
                raise ValueError(
                    f"Invalid case convention value {_rename} for 'rename' in : {attribute!r}."
                ) from None
        else:
            return _rename
    else:
        raise RuntimeError("Unreachable code")



# SPDX-License-Identifier: MIT
from __future__ import annotations

import types

# Copied from CPYTHON/Lib/dataclasses.py (near line 225)
#
# Atomic immutable types which don't require any recursive handling and for which deepcopy
# returns the same object. We can provide a fast-path for these types in asdict and astuple.
_ATOMIC_TYPES = frozenset(
    {
        # Common JSON Serializable types
        types.NoneType,
        bool,
        int,
        float,
        str,
        # Other common types
        complex,
        bytes,
        # Other types that are also unaffected by deepcopy
        types.EllipsisType,
        types.NotImplementedType,
        types.CodeType,
        types.BuiltinFunctionType,
        types.FunctionType,
        type,
        range,
        property,
    }
)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import Any, Callable, Hashable, Mapping
from xattrs._typing import T
from xattrs.typing import StructAs

from copy import copy as shallowcopy
from copy import deepcopy
from functools import partial

from xattrs._metadata import _gen_field_filter, _gen_field_key_serializer
from xattrs._serde import gen_serializer_helpers
from xattrs._types import _ATOMIC_TYPES
from xattrs._uni import _fields, _is_data_class_like_instance
from xattrs.converters import identity

__all__ = (
    "asdict",
    "asdict_shallow",
    "astree",
    "astree_shallow",
    "astuple",
    "astuple_shallow",
)


_AS_DICT = "__attrs_asdict__"
_AS_TUPLE = "__attrs_astuple__"
_FROM_DICT = "__attrs_fromdict__"
_FROM_TUPLE = "__attrs_fromtuple__"


def _as_primitive(
    inst: Any,
    *,
    factory: Callable[[], T],
    filter=None,
    key_serializer=None,
    value_serializer=None,
    copy=deepcopy,
) -> T:
    if isinstance(inst, type):
        raise ValueError("Must be an instance")

    cls = type(inst)

    if cls in _ATOMIC_TYPES:
        return inst
    elif _is_data_class_like_instance(inst):
        ...
        # inst_fields = _fields(inst)
        # inst_serde_params = _get_serde(inst) or {}

        # cls_key_serializer = key_serializer
        # cls_value_serializer = value_serializer

    raise NotImplementedError


def asdict(
    inst: Any,
    *,
    dict_factory: type[Mapping] = dict,
    filter=None,
    key_serializer=None,
    value_serializer=None,
    copy=deepcopy,
) -> Mapping[Hashable, Any]:
    """
    Return the fields of a dataclass or attrs instance as a new dictionary mapping
    field names to field values.
    """
    if isinstance(inst, type):
        raise TypeError("Must be an instance")
    return _asdict_inner(
        inst, dict_factory, filter, key_serializer, value_serializer, copy
    )


def _asdict_inner(  # noqa: PLR0911
    inst: Any, dict_factory, filter_, key_serializer, value_serializer, copy
):
    cls = type(inst)
    args = (filter_, key_serializer, value_serializer, copy)

    if cls in _ATOMIC_TYPES:
        return inst
    elif _is_data_class_like_instance(inst):
        # fast path for the common case of a dataclass / attrs instance
        inst_filter, inst_key_ser, inst_val_ser = gen_serializer_helpers(inst)
        _filter = inst_filter or filter_
        _key_ser = inst_key_ser or key_serializer
        _val_ser = inst_val_ser or value_serializer

        if hasattr(inst, _AS_DICT):
            _ks = _key_ser or identity
            return dict_factory(
                (_ks(k), _asdict_inner(v, dict_factory, *args))
                for k, v in getattr(inst, _AS_DICT)().items()
            )
        else:
            pairs = (
                (
                    _gen_field_key_serializer(f, _key_ser)(f.name),
                    _asdict_inner(getattr(inst, f.name), dict_factory, *args),
                )
                for f in _fields(inst)
                if _gen_field_filter(f, _filter)(f, getattr(inst, f.name))
            )
            return dict_factory(pairs)
    elif isinstance(inst, tuple) and hasattr(inst, "_fields"):
        # instance is a namedtuple.
        # keep namedtuple instances as they are, then recurse into their fields.
        return cls(*(_asdict_inner(v, dict_factory, *args) for v in inst))
    elif isinstance(inst, (list, tuple)):
        return cls(_asdict_inner(v, dict_factory, *args) for v in inst)
    elif isinstance(inst, dict):
        if hasattr(cls, "default_factory"):
            # inst is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = cls.default_factory  # type: ignore
            for k, v in inst.items():
                result[_asdict_inner(k, dict_factory, *args)] = _asdict_inner(
                    v, dict_factory, *args
                )
            return result
        return cls(
            (
                _asdict_inner(k, dict_factory, *args),
                _asdict_inner(v, dict_factory, *args),
            )
            for k, v in inst.items()
        )
    else:
        return copy(inst)


def astuple(
    inst,
    *,
    tuple_factory=tuple,
    key_serializer=None,
    value_serializer=None,
    copy=deepcopy,
):
    """
    Return the fields of a dataclass or attrs instance as a new tuple of field values
    """
    return _astuple_inner(inst, tuple_factory, key_serializer, value_serializer, copy)


def _astuple_inner(  # noqa: PLR0911
    inst, tuple_factory, key_serializer, value_serializer, copy
):
    args = (key_serializer, value_serializer, copy)
    cls = type(inst)

    if cls in _ATOMIC_TYPES:
        return inst
    elif _is_data_class_like_instance(inst):
        result = []
        for f in _fields(inst):
            value = _astuple_inner(getattr(inst, f.name), tuple_factory, *args)
            result.append(value)
        return tuple_factory(result)
    elif isinstance(inst, tuple) and hasattr(inst, "_fields"):
        # instance is a namedtuple.
        # keep namedtuple instances as they are, then recurse into their fields.
        return cls(*[_astuple_inner(v, tuple_factory, *args) for v in inst])
    elif isinstance(inst, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return cls(_astuple_inner(v, tuple_factory, *args) for v in inst)
    elif isinstance(inst, dict):
        dict_cls = cls
        if hasattr(dict_cls, "default_factory"):
            # obj is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = dict_cls.default_factory  # type: ignore
            for k, v in inst.items():
                result[_astuple_inner(k, tuple_factory, *args)] = _astuple_inner(
                    v, tuple_factory, *args
                )
            return result
        return dict_cls(
            (
                _astuple_inner(k, tuple_factory, *args),
                _astuple_inner(v, tuple_factory, *args),
            )
            for k, v in inst.items()
        )
    else:
        return copy(inst)


def astree(
    inst,
    *,
    tuple_factory=tuple,
    key_serializer=None,
    value_serializer=None,
    copy=deepcopy,
):
    """
    Return the fields of a dataclass or attrs instance as a new tuple of field values
    """
    return _astree_inner(inst, tuple_factory, key_serializer, value_serializer, copy)


def _astree_inner(inst, tuple_factory, key_serializer, value_serializer, copy):
    args = (key_serializer, value_serializer, copy)
    cls = type(inst)

    if cls in _ATOMIC_TYPES:
        return inst
    elif _is_data_class_like_instance(inst):
        result = []
        for f in _fields(inst):
            value = _astree_inner(getattr(inst, f.name), tuple_factory, *args)
            result.append(value)
        return tuple_factory(result)
    elif isinstance(inst, tuple) and hasattr(inst, "_fields"):
        # instance is a namedtuple.
        # keep namedtuple instances as they are, then recurse into their fields.
        return cls(*[_astree_inner(v, tuple_factory, *args) for v in inst])
    elif isinstance(inst, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return cls(_astree_inner(v, tuple_factory, *args) for v in inst)
    elif isinstance(inst, Mapping):
        return tuple(
            (
                _astree_inner(k, tuple_factory, *args),
                _astree_inner(v, tuple_factory, *args),
            )
            for k, v in inst.items()
        )
    else:
        return copy(inst)


asdict_shallow = partial(asdict, copy=shallowcopy)
astuple_shallow = partial(astuple, copy=shallowcopy)
astree_shallow = partial(astree, copy=shallowcopy)

_AS_FUNCS_MAPPING: dict[StructAs, Callable] = {
    "dict": _asdict_inner,
    "tuple": _astuple_inner,
    "tree": _astree_inner,
}



from __future__ import annotations

from typing import Any

from collections import OrderedDict
from dataclasses import dataclass

import pytest
from attrs import frozen
from hypothesis import given
from hypothesis import strategies as st

from xattrs._struct_funcs import asdict

MAPPING_TYPES = (dict, OrderedDict)
SEQUENCE_TYPES = (list, tuple)


@pytest.fixture(scope="class")
def A():
    """
    Return a attrs class
    """

    @frozen
    class A:
        x: Any
        y: Any

    return A


@pytest.fixture(scope="class")
def D():
    """
    Return a dataclass class
    """

    @dataclass
    class D:
        x: Any
        y: Any

    return D


class TestAsDict:
    """
    Tests for `asdict`.
    """

    @given(dict_factory=st.sampled_from(MAPPING_TYPES))
    def test_attrs(self, A, dict_factory):
        acutal = asdict(A(A(1, 2), A(3, 4)), dict_factory=dict_factory)
        expected = {"x": {"x": 1, "y": 2}, "y": {"x": 3, "y": 4}}
        assert acutal == expected

    @given(dict_factory=st.sampled_from(MAPPING_TYPES))
    def test_dataclass(self, D, dict_factory):
        actual = {"x": {"x": "1", "y": "2"}, "y": {"x": 3, "y": 4}}
        expected = asdict(D(D("1", "2"), D(3, 4)), dict_factory=dict_factory)
        assert actual == expected

    @given(dict_factory=st.sampled_from(MAPPING_TYPES))
    def test_mixed_attrs_dataclass(self, A, D, dict_factory):
        actual = {"x": {"x": "1", "y": "2"}, "y": {"x": 3, "y": 4}}
        expected = asdict(A(A("1", "2"), D(3, 4)), dict_factory=dict_factory)
        assert actual == expected

        actual = {"x": {"x": "1", "y": "2"}, "y": {"x": 3, "y": 4}}
        expected = asdict(D(A("1", "2"), D(3, 4)), dict_factory=dict_factory)
        assert actual == expected

    @given(seq=st.sampled_from(SEQUENCE_TYPES))
    def test_lists_tuples(self, A, D, seq):
        """
        list / tuple.
        """
        actual = asdict(D(1, seq([A(2, 3), D(4, 5), "a"])))
        expected = {
            "x": 1,
            "y": seq(({"x": 2, "y": 3}, {"x": 4, "y": 5}, "a")),
        }
        assert actual == expected

        actual = asdict(A(1, seq((A(2, 3), D(4, 5), D("6", "7")))))
        expected = {
            "x": 1,
            "y": seq([{"x": 2, "y": 3}, {"x": 4, "y": 5}, {"x": "6", "y": "7"}]),
        }



from __future__ import annotations

from typing import Any

from dataclasses import dataclass
from dataclasses import field as dataclass_field

import pytest
from attr import field as attrs_filed
from attrs import define
from attrs import field as attrs_field

from xattrs._metadata import _gen_field_filter, _gen_field_key_serializer, _Metadata
from xattrs._uni import _fields
from xattrs.converters import to_camel, to_const, to_kebab
from xattrs.filters import exclude_if_default, exclude_if_false, keep_include


@pytest.fixture(scope="class")
def make_field_with_filter():
    def gen(field_kwargs, init_kwargs, metadata: _Metadata):  # type: ignore[type-arg]

        @define
        class A:
            x: Any = attrs_filed(**field_kwargs) | metadata

        @dataclass
        class D:
            x: Any = dataclass_field(**field_kwargs) | metadata

        return A(**init_kwargs), D(**init_kwargs)

    return gen


@pytest.fixture(scope="class")
def make_field_with_alias_name():
    def gen(field_kwargs, init_kwargs, metadata: _Metadata):  # type: ignore[type-arg]

        @define
        class A:
            snake_name: Any = attrs_filed(**field_kwargs) | metadata

        @dataclass
        class D:
            snake_name: Any = dataclass_field(**field_kwargs) | metadata

        return A(**init_kwargs), D(**init_kwargs)

    return gen


class TestMetadata:
    @pytest.mark.parametrize(
        ("deco_func", "field_func"),
        [(define, attrs_field), (dataclass, dataclass_field)],
    )
    @pytest.mark.parametrize(
        "field_kwargs",
        [
            {"default": 3},
            {"metadata": {"exclude": False}},
            {"init": False, "metadata": {"name": "whatever", "exclude": True}},
        ],
    )
    @pytest.mark.parametrize(
        "meta_kwargs",
        [
            {"name": "int_field"},
            {"exclude": True},
            {"converter_to": str},
            {"name": "int_field", "exclude": True, "converter_to": str},
        ],
    )
    def test_metadata__ror_with_attribute(
        self, deco_func, field_func, field_kwargs, meta_kwargs
    ):
        created_field = field_func(**field_kwargs)
        new_field = created_field | _Metadata(**meta_kwargs)
        assert type(new_field) == type(created_field)
        for k, v in meta_kwargs.items():
            assert new_field.metadata.get(k) == v

        @deco_func
        class C:
            x: int = field_func(**field_kwargs) | _Metadata(**meta_kwargs)

        for f in _fields(C):  # pyright: ignore[reportCallIssue,reportArgumentType]
            for k, v in meta_kwargs.items():
                assert f.metadata.get(k) == v

    @pytest.mark.parametrize(
        "meta_kwargs",
        [
            {"exclude": True, "exclude_if": lambda x: x == 3},
            {"exclude": False, "exclude_if_default": True},
            {"exclude": False, "exclude_if_false": True},
            {"exclude": True, "exclude_if_default": False},
            {"exclude_if": lambda x: x == 3, "exclude_if_false": False},
        ],
    )
    def test_metadata__conflict_conf(self, meta_kwargs):
        error_msg = "Only one of 'exclude', 'exclude_if', 'exclude_if_default', 'exclude_if_false' values can be set."
        with pytest.raises(ValueError, match=error_msg):
            _Metadata(**meta_kwargs)


class TestGenFieldFilter:

    @pytest.mark.parametrize(
        ("field_kwargs", "init_kwargs"),
        [
            ({"default": 100}, {"x": 100}),
            ({"default": None}, {"x": 100}),
            ({"default": None}, {"x": 200}),
            ({}, {"x": 0}),
            ({"default": 1000}, {"x": 0}),
            ({"default": 355}, {"x": 355}),
            ({"default": 123}, {"x": 200}),
            ({"default": 123}, {"x": 123}),
        ],
    )
    @pytest.mark.parametrize(
        ("filter_conf", "scope_filter", "expected_filter_func"),
        [
            ({"exclude": True}, None, lambda f, v: False),
            ({"exclude": False}, lambda f, v: True, lambda f, v: True),
            ({"exclude": False}, lambda f, v: False, lambda f, v: True),
            (
                {"exclude_if": lambda f, v: v // 2 == 0},
                lambda f, v: False,
                lambda f, v: v // 2 == 0,
            ),
            (
                {"exclude_if": lambda f, v: v // 2 != 0},
                None,
                lambda f, v: v // 2 != 0,
            ),
            ({"exclude_if_default": True}, lambda f, v: True, exclude_if_default),
            (
                {"exclude_if_default": True},
                lambda f, v: False,
                exclude_if_default,
            ),
            (
                {"exclude_if_default": False},
                lambda f, v: False,
                keep_include,
            ),
            ({"exclude_if_default": False}, None, keep_include),
            ({"exclude_if_false": True}, lambda f, v: True, exclude_if_false),
            ({"exclude_if_false": True}, lambda f, v: False, exclude_if_false),
            ({"exclude_if_false": False}, lambda f, v: True, keep_include),
            ({"exclude_if_false": False}, lambda f, v: False, keep_include),
            ({"exclude_if_false": False}, None, keep_include),
            ({}, None, keep_include),
            ({}, lambda f, v: v // 2 == 0, lambda f, v: v // 2 == 0),
        ],
    )
    def test_gen_field_filter__default(
        self,
        make_field_with_filter,
        field_kwargs,
        init_kwargs,
        filter_conf,
        scope_filter,
        expected_filter_func,
    ):
        attrs_inst, dataclass_inst = make_field_with_filter(
            field_kwargs, init_kwargs, _Metadata(**filter_conf)
        )
        for f in _fields(attrs_inst):
            _filter = _gen_field_filter(f, scope_filter)
            val = getattr(attrs_inst, f.name)
            assert _filter(f, val) == expected_filter_func(f, val)
            break

        for f in _fields(dataclass_inst):
            _filter = _gen_field_filter(f, scope_filter)
            val = getattr(dataclass_inst, f.name)
            assert _filter(f, val) == expected_filter_func(f, val)
            break


class TestGenFieldKeySerializer:

    @pytest.mark.parametrize(
        ("rename_conf", "scope_rename_func", "expected_rename_func"),
        [
            ({"name": "x"}, None, lambda s: "x"),
            ({"name": "x"}, lambda s: "aaa", lambda s: "x"),
            ({"rename": "kebab-case"}, lambda s: False, to_kebab),
            ({"rename": "camelCase"}, None, to_camel),
            ({"rename": lambda s: s.upper()}, lambda s: "xx", to_const),
            ({}, lambda s: "xxx", lambda s: "xxx"),
            ({}, None, lambda s: s),
            ({}, to_const, lambda s: s.upper()),
        ],
    )
    def test_gen_field_filter__default(
        self,
        make_field_with_alias_name,
        rename_conf,
        scope_rename_func,
        expected_rename_func,
    ):
        attrs_inst, dataclass_inst = make_field_with_alias_name(
            {}, {"snake_name": None}, _Metadata(**rename_conf)
        )
        for f in _fields(attrs_inst):
            _filter = _gen_field_key_serializer(f, scope_rename_func)
            assert _filter(f.name) == expected_rename_func(f.name)
        for f in _fields(dataclass_inst):
            _filter = _gen_field_key_serializer(f, scope_rename_func)
            assert _filter(f.name) == expected_rename_func(f.name)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from collections.abc import Sequence
from xattrs._compat.typing import (
    Annotated,
    Any,
    NewType,
    Optional,
    Type,
    TypedDict,
    TypeGuard,
)

from collections import defaultdict
from enum import Enum

from xattrs._types import _ATOMIC_TYPES


def is_atomic(typ: Any) -> bool:
    return type(typ) in _ATOMIC_TYPES


def is_optional(typ: Type) -> TypeGuard[Optional[Any]]: ...


def is_annotated(typ: Any) -> TypeGuard[Annotated]: ...


def is_sequence(typ: Any) -> TypeGuard[Sequence]: ...


def is_hetero_sequence(typ: Any) -> TypeGuard[Sequence]: ...


def is_tuple(typ: Any) -> TypeGuard[tuple]: ...


def is_hetero_tuple(typ: Any) -> TypeGuard[tuple]: ...


def is_typeddict(typ: Any): ...


def is_defaultdict(typ: Any) -> TypeGuard[dict]: ...


def is_dataclass_transform(typ: Any) -> TypeGuard[Annotated]: ...


def is_enum(typ: Any) -> TypeGuard[Enum]:
    try:
        return issubclass(typ, Enum)
    except TypeError:
        return isinstance(typ, Enum)


def is_new_type(type: Any) -> TypeGuard[NewType]: ...



# SPDX-License-Identifier: MIT
"""Configuration constants for xattrs"""

from __future__ import annotations

METANAME = "xattrs"
NAMESPACE_STRUCT_DICT = f"{METANAME}_asdict"
NAMESPACE_STRUCT_TUPLE = f"{METANAME}_astuple"

# Case Convention

LOWER_CASE = "lowercase"
UPPER_CASE = "uppercase"
CAPITALIZE = "capitalize"

PASCAL_CASE = "PascalCase"
CAMEL_CASE = "camelCase"

SNAKE_CASE = "snake_case"
CONST_CASE = "CONST_CASE"
ADA_CASE = "Ada_Case"

KEBAB_CASE = "kebab-case"
COBOL_CASE = "COBOL-CASE"
TRAIN_CASE = "Train-Case"

# Global specific configuration
deny_unknown_fields = "deny_unknown_fields"

# New type (as new type)
NEW_TYPE = "new_type"

# Field specific configuration
ALIAS = "alias"
ALIASES = "aliases"
"""
Duplicate of the original field, may be repeated to specify multiple aliases
"""


SKIP = "skip"
SKIP_DE = "skip_de"
SKIP_SE = "skip_ser"
SKIP_IF = "skip_if"
SKIP_UNLESS = "skip_unless"
SKIP_IF_EMPTY = "skip_if_empty"


SER_NAME = "ser_alias"
DE_NAME = "de_alias"



# SPDX-License-Identifier: MIT
from __future__ import annotations

import pytest
from hypothesis import given
from hypothesis import strategies as st

from xattrs.converters import (
    to_camel,
    to_const,
    to_kebab,
    to_pascal,
    to_secret,
    to_snake,
)

SNAKE_CASES = ["foo", "foo_bar", "foo_bar_baz"]
KEBAB_CASES = ["foo", "foo-bar", "foo-bar-baz"]
CAMEL_CASES = ["foo", "fooBar", "fooBarBaz"]
PASCAL_CASES = ["Foo", "FooBar", "FooBarBaz"]
CONST_CASES = ["FOO", "FOO_BAR", "FOO_BAR_BAZ"]


@pytest.fixture(scope="module")
def snake_cases():
    return set(SNAKE_CASES)


@pytest.fixture(scope="module")
def kebab_cases():
    return set(KEBAB_CASES)


@pytest.fixture(scope="module")
def const_cases():
    return set(CONST_CASES)


@pytest.fixture(scope="module")
def camel_cases():
    return set(CAMEL_CASES)


@pytest.fixture(scope="module")
def pascal_cases():
    return set(PASCAL_CASES)


class TestStringConverters:
    @given(
        kebal=st.sampled_from(KEBAB_CASES + CONST_CASES + CAMEL_CASES + PASCAL_CASES)
    )
    def test_to_snake(self, kebal, snake_cases):
        assert to_snake(kebal) in snake_cases

    @given(
        value=st.sampled_from(SNAKE_CASES + CONST_CASES + CAMEL_CASES + PASCAL_CASES)
    )
    def test_to_kebab(self, value, kebab_cases):
        assert to_kebab(value) in kebab_cases

    @given(
        value=st.sampled_from(SNAKE_CASES + KEBAB_CASES + CAMEL_CASES + PASCAL_CASES)
    )
    def test_to_const(self, value, const_cases):
        assert to_const(value) in const_cases

    # FIXME: failed for CONST_CASES
    @given(value=st.sampled_from(SNAKE_CASES + KEBAB_CASES + PASCAL_CASES))
    def test_to_camel(self, value, camel_cases):
        assert to_camel(value) in camel_cases

    # FIXME: failed for CONST_CASES
    @given(value=st.sampled_from(SNAKE_CASES + KEBAB_CASES + CAMEL_CASES))
    def test_to_pascal(self, value, pascal_cases):
        assert to_pascal(value) in pascal_cases


@pytest.mark.parametrize(
    ("value", "expected", "extra"),
    [
        ("foo", "******", {}),
        ("foo------bar", "######", {"marker": "#"}),
        ("asdfasdfadfbar", "###", {"marker": "#", "length": 3}),
        ("asdfasdfadfbar", "**************", {"length": None}),
    ],
)
def test_to_secret(value, expected, extra):
    assert to_secret(value, **extra) == expected



# SPDX-License-Identifier: MIT
# ruff: noqa: F822
# pyright: reportUnsupportedDunderAll=false
from __future__ import annotations

from functools import partial

import attrs
import optree
from attrs import (
    NOTHING,
    Attribute,
    AttrsInstance,
    Converter,
    Factory,
    asdict,
    assoc,
    astuple,
    cmp_using,
    converters,
    evolve,
    exceptions,
    field,
    fields,
    fields_dict,
    filters,
    has,
    resolve_types,
    setters,
    validate,
    validators,
)
from attrs import define as _define

from xattrs.constants import NAMESPACE_STRUCT_DICT

# Keep the same with the `attrs` module. (except `make_class`)
# `make_class` is not a full replacement in `xattrs`.
__all__ = [*attrs.__all__]  # type: ignore[reportAttributeAccessIssue]  # noqa: PLE0604


def _attrs_flatten(inst):
    cls = inst.__class__
    fs = fields(cls)
    children = tuple(getattr(inst, f.name) for f in fs)
    return (children, fields_dict(cls), {f.name: "*" for f in fs})


def define(cls=None, **kwargs):
    def wrap(_cls):
        attrs_cls = _define(_cls, **kwargs)

        def _attrs_unflatten(metadata, children):
            return attrs_cls(**{f.name: v for f, v in zip(fields(attrs_cls), children)})

        optree.register_pytree_node(
            attrs_cls, _attrs_flatten, _attrs_unflatten, namespace=NAMESPACE_STRUCT_DICT
        )

        return attrs_cls

    if cls is None:
        return wrap

    return wrap(cls)


mutable = define
frozen = partial(define, frozen=True, on_setattr=None)

__getattr__ = __import__("attrs").__getattr__



from typing import Any, Callable, Mapping, Sequence, TypeVar, overload

import sys

from attr import NOTHING as NOTHING
from attr import Attribute as Attribute
from attr import AttrsInstance as AttrsInstance
from attr import Converter as Converter
from attr import Factory as Factory

# Because we need to type our own stuff, we have to make everything from
# attr explicitly public too.
from attr import __author__ as __author__
from attr import __copyright__ as __copyright__
from attr import __description__ as __description__
from attr import __email__ as __email__
from attr import __license__ as __license__
from attr import __title__ as __title__
from attr import __url__ as __url__
from attr import __version__ as __version__
from attr import __version_info__ as __version_info__
from attr import asdict as asdict
from attr import assoc as assoc
from attr import astuple as astuple
from attr import attrib
from attr import cmp_using as cmp_using
from attr import converters as converters
from attr import evolve as evolve
from attr import exceptions as exceptions
from attr import fields as fields
from attr import fields_dict as fields_dict
from attr import filters as filters
from attr import has as has
from attr import make_class as make_class
from attr import resolve_types as resolve_types
from attr import setters as setters
from attr import validate as validate
from attr import validators as validators

if sys.version_info >= (3, 11):
    from typing import dataclass_transform
else:
    from typing_extensions import dataclass_transform

_T = TypeVar("_T")
_C = TypeVar("_C", bound=type)

_EqOrderType = bool | Callable[[Any], Any]
_ValidatorType = Callable[[Any, "Attribute[_T]", _T], Any]
_ConverterType = Callable[[Any], Any]
_ReprType = Callable[[Any], str]
_ReprArgType = bool | _ReprType
_OnSetAttrType = Callable[[Any, "Attribute[Any]", Any], Any]
_OnSetAttrArgType = _OnSetAttrType | list[_OnSetAttrType] | setters._NoOpType
_FieldTransformer = Callable[[type, list["Attribute[Any]"]], list["Attribute[Any]"]]
# FIXME: in reality, if multiple validators are passed they must be in a list
# or tuple, but those are invariant and so would prevent subtypes of
# _ValidatorType from working when passed in a list or tuple.
_ValidatorArgType = _ValidatorType[_T] | Sequence[_ValidatorType[_T]]

@overload
def field(
    *,
    default: None = ...,
    validator: None = ...,
    repr: _ReprArgType = ...,
    hash: bool | None = ...,
    init: bool = ...,
    metadata: Mapping[Any, Any] | None = ...,
    converter: None = ...,
    factory: None = ...,
    kw_only: bool = ...,
    eq: bool | None = ...,
    order: bool | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    alias: str | None = ...,
    type: type | None = ...,
) -> Any: ...

# This form catches an explicit None or no default and infers the type from the
# other arguments.
@overload
def field(
    *,
    default: None = ...,
    validator: _ValidatorArgType[_T] | None = ...,
    repr: _ReprArgType = ...,
    hash: bool | None = ...,
    init: bool = ...,
    metadata: Mapping[Any, Any] | None = ...,
    converter: _ConverterType | Converter[Any, _T] | None = ...,
    factory: Callable[[], _T] | None = ...,
    kw_only: bool = ...,
    eq: _EqOrderType | None = ...,
    order: _EqOrderType | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    alias: str | None = ...,
    type: type | None = ...,
) -> _T: ...

# This form catches an explicit default argument.
@overload
def field(
    *,
    default: _T,
    validator: _ValidatorArgType[_T] | None = ...,
    repr: _ReprArgType = ...,
    hash: bool | None = ...,
    init: bool = ...,
    metadata: Mapping[Any, Any] | None = ...,
    converter: _ConverterType | Converter[Any, _T] | None = ...,
    factory: Callable[[], _T] | None = ...,
    kw_only: bool = ...,
    eq: _EqOrderType | None = ...,
    order: _EqOrderType | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    alias: str | None = ...,
    type: type | None = ...,
) -> _T: ...

# This form covers type=non-Type: e.g. forward references (str), Any
@overload
def field(
    *,
    default: _T | None = ...,
    validator: _ValidatorArgType[_T] | None = ...,
    repr: _ReprArgType = ...,
    hash: bool | None = ...,
    init: bool = ...,
    metadata: Mapping[Any, Any] | None = ...,
    converter: _ConverterType | Converter[Any, _T] | None = ...,
    factory: Callable[[], _T] | None = ...,
    kw_only: bool = ...,
    eq: _EqOrderType | None = ...,
    order: _EqOrderType | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    alias: str | None = ...,
    type: type | None = ...,
) -> Any: ...
@overload
@dataclass_transform(field_specifiers=(attrib, field))
def define(
    maybe_cls: _C,
    *,
    these: dict[str, Any] | None = ...,
    repr: bool = ...,
    unsafe_hash: bool | None = ...,
    hash: bool | None = ...,
    init: bool = ...,
    slots: bool = ...,
    frozen: bool = ...,
    weakref_slot: bool = ...,
    str: bool = ...,
    auto_attribs: bool = ...,
    kw_only: bool = ...,
    cache_hash: bool = ...,
    auto_exc: bool = ...,
    eq: bool | None = ...,
    order: bool | None = ...,
    auto_detect: bool = ...,
    getstate_setstate: bool | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    field_transformer: _FieldTransformer | None = ...,
    match_args: bool = ...,
) -> _C: ...
@overload
@dataclass_transform(field_specifiers=(attrib, field))
def define(
    maybe_cls: None = ...,
    *,
    these: dict[str, Any] | None = ...,
    repr: bool = ...,
    unsafe_hash: bool | None = ...,
    hash: bool | None = ...,
    init: bool = ...,
    slots: bool = ...,
    frozen: bool = ...,
    weakref_slot: bool = ...,
    str: bool = ...,
    auto_attribs: bool = ...,
    kw_only: bool = ...,
    cache_hash: bool = ...,
    auto_exc: bool = ...,
    eq: bool | None = ...,
    order: bool | None = ...,
    auto_detect: bool = ...,
    getstate_setstate: bool | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    field_transformer: _FieldTransformer | None = ...,
    match_args: bool = ...,
) -> Callable[[_C], _C]: ...

mutable = define

@overload
@dataclass_transform(frozen_default=True, field_specifiers=(attrib, field))
def frozen(
    maybe_cls: _C,
    *,
    these: dict[str, Any] | None = ...,
    repr: bool = ...,
    unsafe_hash: bool | None = ...,
    hash: bool | None = ...,
    init: bool = ...,
    slots: bool = ...,
    frozen: bool = ...,
    weakref_slot: bool = ...,
    str: bool = ...,
    auto_attribs: bool = ...,
    kw_only: bool = ...,
    cache_hash: bool = ...,
    auto_exc: bool = ...,
    eq: bool | None = ...,
    order: bool | None = ...,
    auto_detect: bool = ...,
    getstate_setstate: bool | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    field_transformer: _FieldTransformer | None = ...,
    match_args: bool = ...,
) -> _C: ...
@overload
@dataclass_transform(frozen_default=True, field_specifiers=(attrib, field))
def frozen(
    maybe_cls: None = ...,
    *,
    these: dict[str, Any] | None = ...,
    repr: bool = ...,
    unsafe_hash: bool | None = ...,
    hash: bool | None = ...,
    init: bool = ...,
    slots: bool = ...,
    frozen: bool = ...,
    weakref_slot: bool = ...,
    str: bool = ...,
    auto_attribs: bool = ...,
    kw_only: bool = ...,
    cache_hash: bool = ...,
    auto_exc: bool = ...,
    eq: bool | None = ...,
    order: bool | None = ...,
    auto_detect: bool = ...,
    getstate_setstate: bool | None = ...,
    on_setattr: _OnSetAttrArgType | None = ...,
    field_transformer: _FieldTransformer | None = ...,
    match_args: bool = ...,
) -> Callable[[_C], _C]: ...



from xattrs._compat.typing import Callable, TypeAlias

import pytest

from xattrs import derive

_StrGen: TypeAlias = Callable[..., str]


def trait1(wrapped: _StrGen) -> _StrGen:
    """add postfix "1" to wrapped return"""

    def wrapper(**kwargs):
        return wrapped(**kwargs) + "1"

    return wrapper


def trait2(wrapped: _StrGen) -> _StrGen:
    """add prefix "2" to wrapped return"""

    def wrapper(**kwargs):
        return "2" + wrapped(**kwargs)

    return wrapper


def trait3(wrapped: _StrGen) -> _StrGen:
    """reverse the return of wrapped."""

    def wrapper(**kwargs):
        return wrapped(**kwargs)[::-1]

    return wrapper


@pytest.mark.parametrize(
    ("traits", "expected_decorator"),
    [
        ([trait1], trait1),
        (
            [trait1, trait2, trait3],
            lambda x: trait3(trait2(trait1(x))),
        ),
        (
            [trait3, trait2, trait1],
            lambda x: trait1(trait2(trait3(x))),
        ),
    ],
)
def test_derive__sequence(traits, expected_decorator):
    def wrapped():
        return "hello, world"

    assert derive(*traits)(wrapped)() == expected_decorator(wrapped)()


def test_drive__empty():
    """
    Test derive() with no traits. should return identicial function as wrapped.
    """

    def wrapped(x=None):
        return x

    assert id(derive()(wrapped)) == id(wrapped)
    assert derive()(wrapped)() is None
    assert derive()(wrapped)(111111) == 111111



from __future__ import annotations

from dataclasses import is_dataclass

import pytest

from xattrs._uni import _is_dataclass_class


# TODO: add test cases
def _test__is_dataclass_with_offcial_implmentation(data):
    cls = data if isinstance(data, type) else type(data)
    assert _is_dataclass_class(cls) == is_dataclass(data)



# SPDX-License-Identifier: MIT
# mypy: disable-error-code="empty-body"
from __future__ import annotations

from types import MappingProxyType
from xattrs._compat.typing import TYPE_CHECKING, Any, Callable, Generic
from xattrs._typing import SingleDispatchCallable, T, T_co, T_contra, T_interm, T_proto

from abc import abstractmethod

if TYPE_CHECKING:
    from xattrs._typing import ConstructHook, DeconstructHook, Dispatchable


class AbstractSerializer(Generic[T_contra, T_co]):
    """
    Interface for custom class serializer.

    This protocol is intended to be used for custom class serializer.

    >>> from datetime import datetime
    >>>
    >>> class MySerializer(Serializer):
    ...     @dispatch
    ...     def serialize(self, value: datetime) -> str:
    ...         return value.strftime("%d/%m/%y")
    """

    def __call__(self, value: T_contra, *, inst=None, attr=None, **kwargs) -> T_co: ...


class AbstractDeserializer(Generic[T_contra, T]):
    """
    Interface for custom class deserializer.

    This protocol is intended to be used for custom class deserializer.

    >>> from datetime import datetime
    >>>
    >>> class MyDeserializer(Deserializer):
    ...     @dispatch
    ...     def deserialize(self, cls: type[datetime], value: Any) -> datetime:
    ...         return datetime.strptime(value, "%d/%m/%y")
    """

    def __call__(
        self, value: T_contra, cls: type[T], *, inst=None, attr=None, **kwargs
    ) -> T: ...


class AbstractSerDe(Generic[T, T_proto, T_contra]):
    serializer: AbstractSerializer[T, T_proto]
    deserializer: AbstractDeserializer[T_contra, T_co]

    def serialize(self, value: T, **kwargs) -> T_proto:
        return self.serializer(value, **kwargs)

    def deserialize(self, value: T_contra, cls: type[T_co], **kwargs) -> T_co:
        return self.deserializer(value, cls, **kwargs)


###############################################################################


class AbstractDispatcher(Generic[T]):
    _registry: MappingProxyType[str, SingleDispatchCallable[T]]

    @abstractmethod
    def dispatch(self, cls: Dispatchable[T]) -> Callable[..., T]: ...

    @abstractmethod
    def register(
        self, cls: Dispatchable[T], func: Callable[..., T]
    ) -> Callable[..., T]: ...



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import TYPE_CHECKING, Any
from xattrs._typing import T, T_interm, T_proto

from xattrs.abc import AbstractDeserializer, AbstractSerDe, AbstractSerializer

if TYPE_CHECKING:
    from xattrs._typing import ConstructHook, DeconstructHook, Dispatchable


class BaseSerializer(AbstractSerializer[T_interm, T_proto]):
    """Base serializer."""

    def _dumps(self, obj: Any, **kwargs) -> T_proto:
        return self.encode(self.deconstruct(obj, **kwargs), **kwargs)


class BaseDeserializer(AbstractDeserializer[T_proto, T_interm]):
    """Base deserializer."""

    def decode(self, data: T_proto, **kwargs) -> T_interm:
        raise NotImplementedError

    def construct(self, data: T_interm, obj: type[T], **kwargs) -> T:
        raise NotImplementedError

    def loads(self, data: T_proto, obj: type[T], **kwargs) -> T:
        return self.construct(self.decode(data, **kwargs), obj)


class BaseSerDe(AbstractSerDe[T_interm, T_interm, T_proto]):
    """Base serializer/deserializer."""

    serializer: BaseSerializer[T_interm, T_proto]
    deserializer: BaseDeserializer[T_proto, T_interm]

    def loads(
        self, data: T_proto, obj: type[T], **kwargs
    ) -> T:  # pyright: ignore[reportReturnType]
        self.deserializer.loads(data, obj, **kwargs)

    def dumps(self, obj: Any, **kwargs) -> T_proto:  # pyright: ignore[reportReturnType]
        self.serializer.dumps(obj, **kwargs)



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import Any, TypeGuard, cast, overload
from xattrs._typing import (
    AttrsInstance,
    DataclassInstance,
    DataclassLike,
    _AttrsParams,
    _DataclassParams,
)

from dataclasses import MISSING, Field
from dataclasses import fields as dataclass_fields
from dataclasses import is_dataclass

from attr._make import _CountingAttr
from attrs import NOTHING, Attribute
from attrs import fields as attrs_fields

_ATTRS_ATTRS = "__attrs_attrs__"
_DATACLASS_FIELDS = "__dataclass_fields__"
_DATACLASS_PARAMS = "_dataclass_params__"


def _is_attrs_instance(inst: Any) -> TypeGuard[AttrsInstance]:
    """Return True if the object is an instance of a attrs."""
    return hasattr(type(inst), _ATTRS_ATTRS)


def _is_attrs_class(cls: type) -> TypeGuard[type[AttrsInstance]]:
    """Returns True if object is an attrs class"""
    return hasattr(cls, _ATTRS_ATTRS)


def _is_attrs(obj: Any | type) -> TypeGuard[AttrsInstance | type[AttrsInstance]]:
    """Returns True if object is an attrs or an instance of an attrs."""
    cls = obj if isinstance(obj, type) else type(obj)
    return _is_attrs_class(cls)


def _is_dataclass_instance(inst: Any) -> TypeGuard[DataclassInstance]:
    """Return True if the object is an instance of a dataclass."""
    return hasattr(type(inst), _DATACLASS_FIELDS)


def _is_dataclass_class(cls: type) -> TypeGuard[type[DataclassInstance]]:
    """Returns True if object is a dataclass class"""
    return hasattr(cls, _DATACLASS_FIELDS)


def _is_dataclass(
    obj: Any | type,
) -> TypeGuard[DataclassInstance | type[DataclassInstance]]:
    """Returns True if object is a dataclass or an instance of a dataclass."""
    return is_dataclass(obj)


def _is_data_class_like(obj: Any | type) -> bool:
    """Return True if the object is a dataclass like class or
    a instance of a dataclass like class."""
    return _is_dataclass(obj) or _is_attrs(obj)


def _is_data_class_like_instance(obj: Any) -> bool:
    """Return True if the object is an instance of a dataclass like class."""
    return _is_dataclass_instance(obj) or _is_attrs_instance(obj)


@overload
def _get_params(
    obj: DataclassInstance | type[DataclassInstance],
) -> _DataclassParams: ...


@overload
def _get_params(
    obj: AttrsInstance | type[AttrsInstance],
) -> _AttrsParams: ...


def _get_params(obj: Any) -> _DataclassParams | _AttrsParams:
    """Return the parameters of the dataclass-like instance."""
    cls = obj if isinstance(obj, type) else type(obj)
    if _is_attrs_class(cls):
        raise NotImplementedError("attrs instance is not supported yet.")
    elif is_dataclass(cls):
        return getattr(cls, _DATACLASS_PARAMS)
    else:
        raise TypeError("The object is not a dataclass-like type.")


@overload
def _fields(obj: DataclassInstance | type[DataclassInstance]) -> tuple[Field]: ...


@overload
def _fields(obj: AttrsInstance | type[AttrsInstance]) -> tuple[Attribute]: ...


def _fields(obj: Any) -> tuple[Attribute[Any]] | tuple[Field[Any]]:
    """Return a tuple describing the fields of this dataclass or attrs.

    Accepts a dataclass (attrs) or an instance of one. Tuple elements are of
    type Field (Attribute if attrs).
    """
    if not _is_data_class_like(obj):
        raise TypeError(
            "must be called with a data class like (attrs/dataclass) type or instance"
        )

    cls = obj if isinstance(obj, type) else type(obj)
    if _is_attrs_class(cls):
        # `fields` of attrs does not accept an instance, only a class.
        return attrs_fields(cls)
    else:
        return dataclass_fields(cls)  # pyright: ignore[reportReturnType]


def _is_field_like_instance(inst: Any) -> bool:
    if isinstance(inst, (Field, _CountingAttr, Attribute)):
        return True
    else:
        return False


def _should_impl_dataclass(cls: type) -> bool:
    """Return True if the object should implement the dataclass-like protocol."""
    return not _is_data_class_like(cls)


def _collect_fields(inst: DataclassLike) -> tuple:
    """Collect defined fields if the object has not yet been implemented as
    a dataclass-like class."""
    ...


def _is_frozen(cls: Any) -> bool:
    """Return True if the object is frozen."""
    if _is_attrs_class(cls):
        return not hasattr(cls, "__attrs_own_setattr__")
    elif _is_dataclass_class(cls):
        return cast(_DataclassParams, getattr(cls, _DATACLASS_PARAMS)).frozen
    else:
        raise TypeError("The object is not a dataclass-like instance or class.")



# SPDX-License-Identifier: MIT
from __future__ import annotations

import types
from xattrs._compat.typing import Any, AnyStr, Callable, TypeVar
from xattrs.typing import CaseConvention, DeserializeFunc, SerializeFunc

from datetime import datetime
from functools import partial
from io import StringIO

from ruamel.yaml import YAML

from xattrs._struct_funcs import asdict_shallow
from xattrs.deserializer import Deserializer
from xattrs.serializer import Serializer

__all__ = ("dump", "dumps", "from_yaml", "load", "loads", "to_yaml")

# Supports the following objects and types by default:
#
# +-------------------+---------------+
# | Python            | Yaml          |
# +===================+===============+
# | dict              | map           |
# +-------------------+---------------+
# | list, tuple       | seq           |
# +-------------------+---------------+
# | str               | string        |
# +-------------------+---------------+
# | int               | int           |
# +-------------------+---------------+
# | float             | float         |
# +-------------------+---------------+
# | True              | true          |
# +-------------------+---------------+
# | False             | false         |
# +-------------------+---------------+
# | None              | null          |
# +-------------------+---------------+
#

T = TypeVar("T")

Yamlable = dict | list | tuple | str | int | float | bool | types.NoneType  # type: ignore[type-arg]


def _yaml_dumps(obj: Any, *, _ruamel_yaml: YAML, **kwargs: Any) -> str:
    """Serialize ``obj`` to a JSON-formatted ``str``."""
    with StringIO() as stream:
        _ruamel_yaml.dump(obj, stream, **kwargs)
        return stream.getvalue()


def _yaml_loads(obj: str, *, _ruamel_yaml: YAML, **kwargs: Any) -> str:
    """Serialize ``obj`` to a JSON-formatted ``str``."""
    with StringIO() as stream:
        stream.write(obj)
        stream.flush()
        stream.seek(0)
        data = _ruamel_yaml.load(stream, **kwargs)
        return data


class YamlDeserializer(Deserializer[str, Yamlable]):
    """JSON deserializer."""

    def __init__(self, loads: Callable | None = None):
        if loads is None:
            _yaml = YAML(typ="safe")
            _yaml.default_flow_style = False

            self._loads = partial(_yaml_loads, _ruamel_yaml=_yaml)
        else:
            self._loads = loads

    # @Deserializer.dispatch(datetime)
    def _datetime_from_isoformat(self, value: str, **kw) -> datetime:
        """Convert the value to a Python object."""
        return datetime.fromisoformat(value)

    def loads(self, s: AnyStr, **kw) -> Any: ...
    def load(self, fp: Any, **kw) -> Any: ...

    def from_yaml(
        self,
        cls: type[T],
        s: AnyStr,
        *,
        deserializer: DeserializeFunc | None = None,
        loads: Callable | None = None,
        **kw,
    ) -> T:
        """Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance
        containing a JSON document) to a Python object.
        """
        deserializer = deserializer or self
        loads = loads or self.loads
        return loads(fromdict(s, cls), **kw)


class YamlSerializer(Serializer[Yamlable, str]):
    """YAML serializer."""

    def __init__(
        self,
        case: CaseConvention | None = None,
        exclude_if_default: bool = False,
        global_filter: None = None,
        dumps: Callable | None = None,
    ):
        if dumps is None:
            _yaml = YAML(typ="safe")
            _yaml.default_flow_style = False
            self._dumps = partial(_yaml_dumps, _ruamel_yaml=_yaml)
        else:
            self._dumps = dumps

        self._global_key_serializer = None

    # @Serializer.dispatch(datetime)
    def _datetime_to_isoformat(self, value: datetime, **kw) -> str:
        """Convert the value to an intermediate data types."""
        return value.isoformat()

    def dumps(self, obj: Any, **kw) -> str: ...
    def dump(self, obj: Any, fp: Any, **kw) -> None: ...

    def to_yaml(
        self,
        obj: Any,
        *,
        key_serializer: Callable[[str], str] | None = None,
        value_serializer: SerializeFunc | None = None,
        dumps: Callable | None = None,
        **kwargs: Any,
    ) -> str:
        """Serialize ``obj`` to a JSON-formatted ``str``."""
        value_serializer = value_serializer or self
        dumps = dumps or self._dumps
        return dumps(
            asdict_shallow(
                obj, key_serializer=key_serializer, value_serializer=value_serializer
            ),
            **kwargs,
        )


yaml_serializer = YamlSerializer()
yaml_deserializer = YamlDeserializer()

to_yaml = yaml_serializer.to_yaml
from_yaml = yaml_deserializer.from_yaml
dumps = yaml_serializer.dumps
loads = yaml_deserializer.loads
dump = yaml_serializer.dump
load = yaml_deserializer.load



# SPDX-License-Identifier: MIT
from __future__ import annotations

from xattrs._compat.typing import Any
from xattrs._typing import T

from enum import Enum

from xattrs._guards import is_enum


def _enum_value(inst: Any, cls: type[Enum]):
    """Return the value from constructed enum instance or itself if it's already enum."""
    if is_enum(inst):
        value = inst.value
        if is_enum(value):
            return _enum_value(value, type(value))
        else:
            return value
    else:
        return cls(inst).value


def _identity(value: T) -> T:
    return value



from __future__ import annotations

from textwrap import dedent

from attrs import define, field

from xattrs import asdict, serde
from xattrs.preconf.yaml import to_yaml


def test_yaml__attrs_example():
    @serde(rename="kebab-case")
    @define
    class Person:
        first_name: str
        last_name: str
        age: int

        full_name: str = field(init=False, alias="full_name")

        def __attrs_post_init__(self):
            self.full_name = f"{self.first_name} {self.last_name}"

    data = {"first_name": "John", "last_name": "Lowe", "age": 25}
    person = Person(**data)
    assert (
        repr(person)
        == "Person(first_name='John', last_name='Lowe', age=25, full_name='John Lowe')"
    )

    assert asdict(person) == {
        "first-name": "John",
        "last-name": "Lowe",
        "age": 25,
        "full-name": "John Lowe",
    }

    yaml_string = dedent(
        """
        age: 25
        first-name: John
        full-name: John Lowe
        last-name: Lowe
        """
    )
    assert to_yaml(person).strip() == yaml_string.strip()



# SPDX-License-Identifier: MIT
from __future__ import annotations

from types import MappingProxyType
from xattrs._compat.typing import (
    Any,
    Callable,
    TypedDict,
    dataclass_transform,
    overload,
)
from xattrs._typing import AttrsInstance, DataclassInstance, T
from xattrs.typing import (
    CaseConvention,
    CaseConverter,
    FilterBuiltins,
    FilterCallable,
    KeyConverter,
    StructAs,
    UnknownFields,
)

from dataclasses import dataclass

from attrs import Attribute, fields

from xattrs._uni import _is_frozen
from xattrs.converters import _CASE_CONVERTER_MAPPING
from xattrs.filters import keep_include

_ATTRS_SERDE = "__attrs_serde__"


@dataclass(slots=True)
class SerdeParams:
    rename: CaseConvention | CaseConverter | None = None
    # alias: str | None = None
    # alias_converter: str | CaseConvention | None = None
    kind: StructAs | None = None
    filter: FilterCallable[Any] | None = None
    tag: str | Callable | None = None  # type: ignore[type-arg]
    tag_field: str = "type"
    unknown_fields: UnknownFields | None = None
    value_serializer: None = None
    value_deserializer: None = None
    schema: None = None
    metadata: MappingProxyType | None = None  # type: ignore[type-arg]


@overload
def serde(
    cls: type[T],
    /,
    *,
    rename: CaseConvention | None = None,
    kind: StructAs | None = None,
    filter: FilterCallable[Any] | None = None,
    tag: str | Callable | None = None,  # type: ignore[type-arg]
    tag_field: str = "type",
    unknown_fields: UnknownFields | None = None,
    value_serializer: None = None,
    value_deserializer: None = None,
    schema: None = None,
    metadata: MappingProxyType | None = None,  # type: ignore[type-arg]
) -> type[T]: ...


@overload
def serde(
    cls: Any = None,
    /,
    *,
    rename: CaseConvention | CaseConverter | None = None,
    kind: StructAs | None = None,
    filter: FilterCallable[Any] | None = None,
    tag: str | Callable | None = None,  # type: ignore[type-arg]
    tag_field: str = "type",
    unknown_fields: UnknownFields | None = None,
    value_serializer: None = None,
    value_deserializer: None = None,
    schema: None = None,
    metadata: MappingProxyType | None = None,  # type: ignore[type-arg]
) -> Callable[[type[T]], type[T]]: ...


@dataclass_transform()
def serde(
    cls: type[T] | None = None,
    /,
    *,
    rename: CaseConvention | CaseConverter | None = None,
    kind: StructAs | None = None,
    filter: FilterBuiltins | FilterCallable | None = None,  # type: ignore[type-arg]
    tag: str | Callable | None = None,  # type: ignore[type-arg]
    tag_field: str = "type",
    unknown_fields: UnknownFields | None = None,
    value_serializer: None = None,
    value_deserializer: None = None,
    schema: None = None,
    metadata: MappingProxyType | None = None,  # type: ignore[type-arg]
):
    def wrapper(cls: type[T]) -> type[T]:
        return _process_serde(  # type: ignore[no-any-return]
            cls,
            rename=rename,
            kind=kind,
            filter=filter,
            tag=tag,
            tag_field=tag_field,
            unknown_fields=unknown_fields,
            value_serializer=value_serializer,
            value_deserializer=value_deserializer,
            schema=schema,
            metadata=metadata,
        )

    if cls is None:
        return wrapper

    return wrapper(cls)


def _process_serde(cls: type[T], **kwargs):
    """Process the `serde` decorator."""
    _serde = SerdeParams(**kwargs)
    if _is_frozen(cls):
        raise NotImplementedError("Frozen classes are not supported.")
    else:
        setattr(cls, _ATTRS_SERDE, _serde)
    return cls


# def _get_serde(obj: XattrsInstance) -> SerdeParams:
#     """Get the `SerdeParams` object from class or instance"""
#     cls = obj if isinstance(obj, type) else type(obj)
#     serde: SerdeParams | None = getattr(cls, _ATTRS_SERDE)
#     if serde is None:
#         raise ValueError(f"no serde params found for {cls!r}")
#     return serde


def _maybe_serde(obj: Any) -> SerdeParams | None:
    """Get the `SerdeParams` object from class or instance."""
    cls = obj if isinstance(obj, type) else type(obj)
    return getattr(cls, _ATTRS_SERDE, None)


def get_serde_kind(obj: Any) -> StructAs | None:
    _serde = _maybe_serde(obj)
    return _serde.kind if _serde else None


def gen_cls_filter(params: SerdeParams) -> FilterCallable:  # type: ignore[type-arg]
    return keep_include


def gen_serializer_helpers(
    obj: (
        AttrsInstance
        | DataclassInstance
        | type[AttrsInstance]
        | type[DataclassInstance]
    ),
    /,
    **kwargs,
) -> tuple[FilterCallable | None, KeyConverter | None, Callable | None]:  # type: ignore[type-arg]
    """Create per class level filter, key serializer and value serializer from attrs like instance."""
    serde = _maybe_serde(obj)
    if serde is None:
        return None, None, None

    _filter = serde.filter

    if (rename := serde.rename) is None:
        _key_serializer = None
    elif isinstance(rename, str):
        try:
            _key_serializer = _CASE_CONVERTER_MAPPING[rename]
        except KeyError:
            raise ValueError(f"unknown alias converter: {rename!r}") from None
    else:
        _key_serializer = rename

    if (val_ser := serde.value_serializer) is None:
        _value_serializer = None
    else:
        _value_serializer = val_ser

    return _filter, _key_serializer, _value_serializer
